/* crosswords-misc.c
 *
 * Copyright 2021 Jonathan Blandford <jrb@gnome.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "crosswords-config.h"
#include <locale.h>
#include <glib/gi18n-lib.h>

#include <adwaita.h>

#include "crosswords-misc.h"
#include "grid-layout.h"
#include "play-grid.h"
#include "puzzle-render.h"

GdkPixbuf *
xwd_pixbuf_load_from_gresource (GResource  *resource,
                                const char *resource_path)
{
  GBytes *bytes;
  g_autoptr (GError) error = NULL;
  g_autoptr (GdkPixbufLoader) loader = NULL;
  GdkPixbuf *pixbuf;

  if (resource == NULL)
    bytes = g_resources_lookup_data (resource_path,
                                     G_RESOURCE_LOOKUP_FLAGS_NONE,
                                     NULL);
  else
    bytes = g_resource_lookup_data (resource, resource_path,
                                    G_RESOURCE_LOOKUP_FLAGS_NONE,
                                    NULL);
  loader = gdk_pixbuf_loader_new ();
  gdk_pixbuf_loader_write_bytes (loader, bytes, &error);
  gdk_pixbuf_loader_close (loader, NULL);

  pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
  if (pixbuf)
    g_object_ref (G_OBJECT (pixbuf));

  return pixbuf;
}

IpuzPuzzle *
xwd_load_puzzle_from_resource (GResource    *resource,
                               const gchar  *resource_path,
                               GError      **error)
{
  g_autoptr (GInputStream) stream = NULL;

  if (resource)
    stream = g_resource_open_stream (resource, resource_path, 0, error);
  else
    stream = g_resources_open_stream (resource_path, 0, error);

  if (stream == NULL)
    return NULL;

  return ipuz_puzzle_new_from_stream (stream, NULL, error);
}


GdkTexture *
xwd_thumbnail_puzzle (IpuzPuzzle *puzzle,
                      guint       size)
{
  cairo_surface_t *surface;
  cairo_t *cr;
  RenderViewport viewport = { 0.0, 0.0, size, size};
  LayoutConfig config = {
    .border_size = 1,
    .base_size = 3,
  };
  GBytes *bytes;
  GdkTexture *texture;

  surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, size, size);
  cr = cairo_create (surface);

  render_puzzle_with_config (cr,
                             &viewport,
                             puzzle,
                             render_color_scheme_from_adwaita (),
                             config);
  cairo_surface_flush(surface);

  /* See
   * https://discourse.gnome.org/t/need-a-simpler-method-to-get-a-gdktexture-from-a-cairo-surface-t/7137
   * and `gdk_texture_new_for_surface()` in
   * https://gitlab.gnome.org/GNOME/gtk/-/blob/main/gdk/gdktexture.c */
  bytes = g_bytes_new_with_free_func (cairo_image_surface_get_data (surface),
                                      size * cairo_image_surface_get_stride (surface),
                                      (GDestroyNotify) cairo_surface_destroy,
                                      cairo_surface_reference (surface));
  texture = gdk_memory_texture_new (size,
                                    size,
                                    GDK_MEMORY_DEFAULT,
                                    bytes,
                                    cairo_image_surface_get_stride (surface));

  g_bytes_unref (bytes);
  cairo_destroy (cr);
  cairo_surface_destroy (surface);

  return texture;
}

guint
utf8_n_clusters (const gchar  *text)
{
  if (text == NULL || *text == '\0')
    return 0;

  /* FIXME(i18n): Actually write this thing. I'm faking it to get the code working
   */
  return (guint) g_utf8_strlen (text, -1);
}

/**
 * utf8_get_next_cluster:
 * @text: a valid utf8 string
 * @end: an end pointer to the next utf8 character in text
 *
 * Returns a newly allocated string of the next text cluster in
 * text. #end is set to be the start of the next cluster in the
 * string.
 *
 * Returns: a newly allocated utf8 string
 **/
gchar *
utf8_get_next_cluster (const gchar  *text,
                       const gchar **end)
{
  const gchar *ptr;
  gchar *retval;
  gunichar c;

  if (text == NULL || *text == '\0')
    {
      *end = text;
      return NULL;
    }

  /* Move forward through any spaces or control characters until we find an uppercase character */
  ptr = text;
  c = g_utf8_get_char (ptr);
  while (ptr[0] && g_unichar_isspace (c))
    {
      ptr = g_utf8_next_char (ptr);
      c = g_utf8_get_char (ptr);
    }

  text = ptr;
  if (!text[0])
    {
      *end = text;
      return NULL;
   }

  /* Go through subsequent characters appending any marks */
  do
    {
      ptr = g_utf8_next_char (ptr);
      c = g_utf8_get_char (ptr);
    }
  while (ptr[0] && g_unichar_ismark (c));

  retval = g_strndup (text, ptr - text);
  *end = ptr;

  return retval;
}

static void
zoom_level_css_clear_or_set (GtkWidget   *widget,
                             const gchar *css_prefix,
                             const gchar *css_class,
                             gboolean     set_css)
{
  g_autofree gchar *real_css_class = NULL;
  gboolean has_css;

  if (css_prefix)
    real_css_class = g_strdup_printf ("%s%s", css_prefix, css_class);
  else
    real_css_class = g_strdup (css_class);

  has_css = gtk_widget_has_css_class (widget, real_css_class);

  /* GTK will blindly update the css class regardless of what's
   * set. Since we are calling this every keypress, we only remove/set
   * the css_class if it's different. */
  if (!has_css && set_css)
    gtk_widget_add_css_class (widget, real_css_class);
  else if (has_css && !set_css)
    gtk_widget_remove_css_class (widget, real_css_class);
}

void
set_zoom_level_css_class (GtkWidget   *widget,
                          const gchar *css_prefix,
                          ZoomLevel    zoom_level)
{
  g_return_if_fail (GTK_IS_WIDGET (widget));

  zoom_level_css_clear_or_set (widget, css_prefix, "zoom-xsmall",
                               (zoom_level == ZOOM_XSMALL));
  zoom_level_css_clear_or_set (widget, css_prefix, "zoom-small",
                               (zoom_level == ZOOM_SMALL));
  zoom_level_css_clear_or_set (widget, css_prefix, "zoom-normal",
                               (zoom_level == ZOOM_NORMAL));
  zoom_level_css_clear_or_set (widget, css_prefix, "zoom-large",
                               (zoom_level == ZOOM_LARGE));
  zoom_level_css_clear_or_set (widget, css_prefix, "zoom-xlarge",
                               (zoom_level == ZOOM_XLARGE));
}

RenderColorScheme
render_color_scheme_from_adwaita (void)
{
  return (adw_style_manager_get_dark (adw_style_manager_get_default ()) ?
          RENDER_COLOR_SCHEME_DARK :
          RENDER_COLOR_SCHEME_LIGHT);
}
