/* edit-acrostic-answers.c
 *
 * Copyright 2024 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more answers.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "crosswords-config.h"
#include <glib/gi18n-lib.h>
#include <libipuz/libipuz.h>
#include <adwaita.h>
#include "crosswords-misc.h"
#include "edit-acrostic-answer.h"
#include "edit-acrostic-answers.h"
#include "acrostic-generator.h"
#include "edit-entry-row.h"
#include "acrostic-answer-list.h"
#include "acrostic-answer-list-row.h"


struct _EditAcrosticAnswers
{
  GtkWidget parent_instance;

  GtkBitset *selected;

  AcrosticAnswerList *answers_model;

  GtkWidget *select_all_button;
  GtkWidget *answers_list_sw;
  GtkWidget *answers_word_list_view;
};


static void     edit_acrostic_answers_init           (EditAcrosticAnswers      *self);
static void     edit_acrostic_answers_class_init     (EditAcrosticAnswersClass *klass);
static void     edit_acrostic_answers_dispose        (GObject                  *object);
static void     setup_listitem_cb                    (GtkListItemFactory       *factory,
                                                      GtkListItem              *list_item);
static void     bind_listitem_cb                     (GtkListItemFactory       *factory,
                                                      GtkListItem              *list_item);
static void     edit_acrostic_answers_select_all_toggled_cb (EditAcrosticAnswers *self);


G_DEFINE_TYPE (EditAcrosticAnswers, edit_acrostic_answers, GTK_TYPE_WIDGET);


static void
edit_acrostic_answers_init (EditAcrosticAnswers *self)
{
  GtkSelectionModel *selection_model;
  g_autoptr (GtkListItemFactory) factory = NULL;

  gtk_widget_init_template (GTK_WIDGET (self));

  self->answers_model = acrostic_answer_list_new ();
  self->selected = gtk_bitset_new_empty ();

  selection_model =
    GTK_SELECTION_MODEL (gtk_no_selection_new (G_LIST_MODEL (self->answers_model)));

  factory = gtk_signal_list_item_factory_new ();

  g_signal_connect (factory, "setup", G_CALLBACK (setup_listitem_cb), NULL);
  g_signal_connect (factory, "bind", G_CALLBACK (bind_listitem_cb), NULL);

  g_object_set (self->answers_word_list_view,
                "model", selection_model,
                "factory", factory,
                NULL);
}

static void
edit_acrostic_answers_class_init (EditAcrosticAnswersClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->dispose = edit_acrostic_answers_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-acrostic-answers.ui");

  gtk_widget_class_bind_template_child (widget_class, EditAcrosticAnswers, select_all_button);
  gtk_widget_class_bind_template_child (widget_class, EditAcrosticAnswers, answers_list_sw);
  gtk_widget_class_bind_template_child (widget_class, EditAcrosticAnswers, answers_word_list_view);

  gtk_widget_class_bind_template_callback (widget_class, edit_acrostic_answers_select_all_toggled_cb);

  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
}

static void
edit_acrostic_answers_dispose (GObject *object)
{
  EditAcrosticAnswers *self;
  GtkWidget *child;

  self = EDIT_ACROSTIC_ANSWERS (object);
  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  g_clear_pointer (&self->selected, gtk_bitset_unref);

  G_OBJECT_CLASS (edit_acrostic_answers_parent_class)->dispose (object);
}

static void
setup_listitem_cb (GtkListItemFactory       *factory,
                   GtkListItem              *list_item)
{
  GtkWidget *row;

  row = edit_acrostic_answer_new (NULL);

  gtk_list_item_set_child (list_item, row);
}

static void
answer_selection_toggled (GtkWidget   *answer,
                          gboolean     active,
                          GtkListItem *list_item)
{
  EditAcrosticAnswers *self;
  GtkWidget *parent;
  guint index;

  index = gtk_list_item_get_position (list_item);

  parent = answer;
  do {
    parent = gtk_widget_get_parent (parent);
  } while (parent != NULL
           && !EDIT_IS_ACROSTIC_ANSWERS (parent));

  /* FIXME: We have a spurious signal from adding the row before the
   * widget is fully parented due to signal connection order. We just
   * exit the callback in that case. The real fix is a deeper refactor.
   */
  if (parent == NULL)
    return;

  self = EDIT_ACROSTIC_ANSWERS (parent);

  if (active)
    gtk_bitset_add (self->selected, index);
  else
    gtk_bitset_remove (self->selected, index);
}

static void
row_changed_cb (AcrosticAnswerListRow *row,
                GtkWidget             *acrostic_answer)
{
  const char *answer;
  gboolean selected;

  answer = acrostic_answer_list_row_get_answer (row);
  selected = acrostic_answer_list_row_get_selected (row);

  edit_acrostic_answer_set_str (EDIT_ACROSTIC_ANSWER (acrostic_answer),
                                answer);
  edit_acrostic_answer_set_selected (EDIT_ACROSTIC_ANSWER (acrostic_answer),
                                     selected);
}

static void
bind_listitem_cb (GtkListItemFactory       *factory,
                  GtkListItem              *list_item)
{
  GtkWidget *acrostic_answer;
  AcrosticAnswerListRow *row;

  acrostic_answer = gtk_list_item_get_child (list_item);
  row = gtk_list_item_get_item (list_item);

  g_signal_connect (acrostic_answer, "answer-selection-toggled",
                    G_CALLBACK (answer_selection_toggled), list_item);
  g_signal_connect (row, "changed",
                    G_CALLBACK (row_changed_cb), acrostic_answer);

  row_changed_cb (row, acrostic_answer);
}

static void
edit_acrostic_answers_select_all_toggled_cb (EditAcrosticAnswers *self)
{
  gboolean active;

  active = gtk_check_button_get_active (GTK_CHECK_BUTTON (self->select_all_button));

  acrostic_answer_list_set_all_selected (self->answers_model, active);
}


static void
clear_answers (EditAcrosticAnswers *self)
{
  acrostic_answer_list_clear (self->answers_model);
  gtk_check_button_set_active (GTK_CHECK_BUTTON (self->select_all_button), FALSE);
  gtk_bitset_remove_all (self->selected);
}

static void
update_answers_from_puzzle (EditAcrosticAnswers *self,
                            GridState           *state)
{
  guint n_answers;
  IpuzClueId clue_id = {
    .direction = IPUZ_CLUE_DIRECTION_CLUES,
  };

  /* FIXME(optimization): Merge this like in play_clue_list_update */
  clear_answers (self);

  n_answers = ipuz_clues_get_n_clues (IPUZ_CLUES (state->xword),
                                      IPUZ_CLUE_DIRECTION_CLUES);

  for (clue_id.index = 0; clue_id.index < n_answers; clue_id.index++)
    {
      g_autofree gchar *answer = NULL;

      answer = ipuz_clues_get_clue_string_by_id (IPUZ_CLUES (state->xword), &clue_id);
      acrostic_answer_list_append (self->answers_model, answer, FALSE);
    }
}


static void
update_answers_from_runner (EditAcrosticAnswers *self,
                            GridState           *grid_state,
                            TaskRunner          *runner,
                            guint                index)
{
  guint n_answers;
  GArray *answers = NULL;
  IpuzClueId clue_id = {
    .direction = IPUZ_CLUE_DIRECTION_CLUES,
  };

  /* FIXME(optimization): Merge this like in play_clue_list_update */
  clear_answers (self);

  n_answers = ipuz_clues_get_n_clues (IPUZ_CLUES (grid_state->xword),
                                      IPUZ_CLUE_DIRECTION_CLUES);
  answers = (GArray *) task_runner_get_result (runner, index);

  for (clue_id.index = 0; clue_id.index < n_answers; clue_id.index++)
    {
      const gchar *autosolve_answer = NULL;

      if (answers)
        autosolve_answer = g_array_index (answers, gchar *, clue_id.index);

      if (autosolve_answer)
        {
          acrostic_answer_list_append (self->answers_model, autosolve_answer, FALSE);
        }
      else
        {
          g_autofree gchar *answer = NULL;

          answer = ipuz_clues_get_clue_string_by_id (IPUZ_CLUES (grid_state->xword), &clue_id);
          acrostic_answer_list_append (self->answers_model, answer, FALSE);
        }
    }
}

void
edit_acrostic_answers_update (EditAcrosticAnswers *self,
                              GridState           *state,
                              TaskRunner          *runner,
                              guint                index)
{
  g_return_if_fail (EDIT_IS_ACROSTIC_ANSWERS (self));

  if (! IPUZ_IS_ACROSTIC (state->xword))
    return;


  if (task_runner_get_state (runner) == TASK_RUNNER_READY)
    update_answers_from_puzzle (self, state);
  else
    update_answers_from_runner (self, state, runner, index);

}

GtkBitset *
edit_acrostic_answers_get_selected (EditAcrosticAnswers *self)
{
  g_return_val_if_fail (EDIT_IS_ACROSTIC_ANSWERS (self), NULL);

  return self->selected;
}
