/* edit-acrostic-details.c
 *
 * Copyright 2024 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */

#include "crosswords-config.h"
#include <glib/gi18n-lib.h>
#include <libipuz/libipuz.h>
#include <adwaita.h>
#include "crosswords-misc.h"
#include "edit-acrostic-answer.h"
#include "edit-acrostic-details.h"
#include "acrostic-generator.h"
#include "edit-entry-row.h"


enum
{
  NORMALIZED_QUOTE_STR_CHANGED,
  NORMALIZED_SOURCE_STR_CHANGED,
  N_SIGNALS
};

static guint obj_signals[N_SIGNALS] = { 0 };


struct _EditAcrosticDetails
{
  GtkWidget parent_instance;

  GtkWidget *quote_row;
  GtkWidget *source_row;
};


static void     edit_acrostic_details_init           (EditAcrosticDetails      *self);
static void     edit_acrostic_details_class_init     (EditAcrosticDetailsClass *klass);
static void     edit_acrostic_details_dispose        (GObject                  *object);
static void     quote_row_applied_cb                 (EditAcrosticDetails      *self,
                                                      EditEntryRow             *entry_row);
static void     source_row_applied_cb                (EditAcrosticDetails      *self,
                                                      EditEntryRow             *entry_row);


G_DEFINE_TYPE (EditAcrosticDetails, edit_acrostic_details, GTK_TYPE_WIDGET);


static void
edit_acrostic_details_init (EditAcrosticDetails *self)
{
  gtk_widget_init_template (GTK_WIDGET (self));
}

static void
edit_acrostic_details_class_init (EditAcrosticDetailsClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->dispose = edit_acrostic_details_dispose;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-acrostic-details.ui");

  gtk_widget_class_bind_template_child (widget_class, EditAcrosticDetails, quote_row);
  gtk_widget_class_bind_template_child (widget_class, EditAcrosticDetails, source_row);

  gtk_widget_class_bind_template_callback (widget_class, quote_row_applied_cb);
  gtk_widget_class_bind_template_callback (widget_class, source_row_applied_cb);

  obj_signals [NORMALIZED_QUOTE_STR_CHANGED] =
    g_signal_new ("normalized-quote-str-changed",
                  EDIT_TYPE_ACROSTIC_DETAILS,
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
                  0,
                  NULL, NULL,
                  NULL,
                  G_TYPE_NONE, 1,
                  G_TYPE_STRING);

  obj_signals [NORMALIZED_SOURCE_STR_CHANGED] =
    g_signal_new ("normalized-source-str-changed",
                  EDIT_TYPE_ACROSTIC_DETAILS,
                  G_SIGNAL_RUN_FIRST | G_SIGNAL_ACTION,
                  0,
                  NULL, NULL,
                  NULL,
                  G_TYPE_NONE, 1,
                  G_TYPE_STRING);

  gtk_widget_class_set_layout_manager_type (widget_class, GTK_TYPE_BIN_LAYOUT);
}

static void
edit_acrostic_details_dispose (GObject *object)
{
  GtkWidget *child;

  while ((child = gtk_widget_get_first_child (GTK_WIDGET (object))))
    gtk_widget_unparent (child);

  G_OBJECT_CLASS (edit_acrostic_details_parent_class)->dispose (object);
}

static void
update_warnings (EditAcrosticDetails *self)
{
  g_autoptr (GError) error = NULL;
  g_autoptr (AcrosticGenerator) acrostic_generator = NULL;
  const gchar *quote_str, *source_str;

  quote_str = gtk_editable_get_text (GTK_EDITABLE (self->quote_row));
  source_str = gtk_editable_get_text (GTK_EDITABLE (self->source_row));

  acrostic_generator =  acrostic_generator_new ();

  if (acrostic_generator_set_text (acrostic_generator, quote_str, source_str, &error))
    {
      edit_entry_row_clear_error (EDIT_ENTRY_ROW (self->source_row));
    }
  else
    {
      edit_entry_row_set_error (EDIT_ENTRY_ROW (self->source_row), error->message);
    }
}

static void
quote_row_applied_cb (EditAcrosticDetails *self,
                      EditEntryRow        *entry_row)
{
  const gchar *quote_str;

  quote_str = gtk_editable_get_text (GTK_EDITABLE (self->quote_row));

  update_warnings (self);
  g_signal_emit (self, obj_signals[NORMALIZED_QUOTE_STR_CHANGED], 0, quote_str);
}

static void
source_row_applied_cb (EditAcrosticDetails *self,
                       EditEntryRow        *entry_row)
{
  const gchar *source_str;

  source_str = gtk_editable_get_text (GTK_EDITABLE (self->source_row));

  update_warnings (self);
  g_signal_emit (self, obj_signals[NORMALIZED_SOURCE_STR_CHANGED], 0, source_str);
}

void
edit_acrostic_details_commit_changes (EditAcrosticDetails *self)
{
  edit_entry_row_commit (EDIT_ENTRY_ROW (self->quote_row));
  edit_entry_row_commit (EDIT_ENTRY_ROW (self->source_row));
}

void
edit_acrostic_details_update (EditAcrosticDetails *self,
                              GridState           *grid_state)
{
  const gchar *quote_str, *source_str;

  if (! IPUZ_IS_ACROSTIC (grid_state->xword))
    return;

  quote_str = ipuz_acrostic_get_quote (IPUZ_ACROSTIC (grid_state->xword));
  source_str = ipuz_acrostic_get_source (IPUZ_ACROSTIC (grid_state->xword));

  /* Update quote row */
  edit_entry_row_set_text (EDIT_ENTRY_ROW (self->quote_row), quote_str);
  edit_entry_row_commit (EDIT_ENTRY_ROW (self->quote_row));

  /* Update source row */
  edit_entry_row_set_text (EDIT_ENTRY_ROW (self->source_row), source_str);
  edit_entry_row_commit (EDIT_ENTRY_ROW (self->source_row));
}

