/* edit-color-row.c
 *
 * Copyright 2024 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */
/* This is a cut-and-paste of edit-shapebg-row.c
 * If we use this pattern another time, we should make a base class
 * for it.
 */


#include "crosswords-config.h"
#include <glib/gi18n-lib.h>
#include "color-palette.h"
#include "edit-color-row.h"
#include "edit-color-swatch.h"


enum
{
  PROP_0,
  PROP_CURRENT_COLOR,
  N_PROPS
};

enum
{
  CHANGED,
  CLEARED,
  N_SIGNALS
};

static GParamSpec *obj_props[N_PROPS] = {NULL, };
static guint obj_signals[N_SIGNALS] = { 0, };


static GdkRGBA style_color_palette[] =
{
  { PALETTE (BLUE_1) },
  { PALETTE (BLUE_2) },
  { PALETTE (BLUE_3) },
  { PALETTE (BLUE_4) },
  { PALETTE (BLUE_5) },

  { PALETTE (GREEN_1) },
  { PALETTE (GREEN_2) },
  { PALETTE (GREEN_3) },
  { PALETTE (GREEN_4) },
  { PALETTE (GREEN_5) },

  { PALETTE (YELLOW_1) },
  { PALETTE (YELLOW_2) },
  { PALETTE (YELLOW_3) },
  { PALETTE (YELLOW_4) },
  { PALETTE (YELLOW_5) },

  { PALETTE (ORANGE_1) },
  { PALETTE (ORANGE_2) },
  { PALETTE (ORANGE_3) },
  { PALETTE (ORANGE_4) },
  { PALETTE (ORANGE_5) },

  { PALETTE (RED_1) },
  { PALETTE (RED_2) },
  { PALETTE (RED_3) },
  { PALETTE (RED_4) },
  { PALETTE (RED_5) },

  { PALETTE (PURPLE_1) },
  { PALETTE (PURPLE_2) },
  { PALETTE (PURPLE_3) },
  { PALETTE (PURPLE_4) },
  { PALETTE (PURPLE_5) },

  { PALETTE (BROWN_1) },
  { PALETTE (BROWN_2) },
  { PALETTE (BROWN_3) },
  { PALETTE (BROWN_4) },
  { PALETTE (BROWN_5) },

  { PALETTE (LIGHT_1) },
  { PALETTE (LIGHT_2) },
  { PALETTE (LIGHT_3) },
  { PALETTE (LIGHT_4) },
  { PALETTE (LIGHT_5) },

  { PALETTE (DARK_1) },
  { PALETTE (DARK_2) },
  { PALETTE (DARK_3) },
  { PALETTE (DARK_4) },
  { PALETTE (DARK_5) },
};
#define STYLE_COLOR_PALETTE_STRIDE 5

struct _EditColorRow
{
  AdwActionRow parent_instance;

  GtkWidget *color_popover;
  GtkWidget *palette_box;
  GtkWidget *clear_button;
  GArray *color_buttons;

  gboolean color_set;
  GdkRGBA current_color;

  GtkWidget *swatches [G_N_ELEMENTS (style_color_palette)];
  gulong style_recalculate_handler;
};


G_DEFINE_FINAL_TYPE (EditColorRow, edit_color_row, ADW_TYPE_ACTION_ROW);


static void edit_color_row_init         (EditColorRow      *self);
static void edit_color_row_class_init   (EditColorRowClass *klass);
static void edit_color_row_set_property (GObject           *object,
                                         guint              prop_id,
                                         const GValue      *value,
                                         GParamSpec        *psec);
static void edit_color_row_get_property (GObject           *object,
                                         guint              prop_id,
                                         GValue            *value,
                                         GParamSpec        *psec);
static void edit_color_row_finalize     (GObject           *object);
static void edit_color_row_activate     (AdwActionRow      *action_row);
static void notify_popover_visible_cb   (EditColorRow      *self);
static void clear_button_cb             (EditColorRow      *self);
static void update_buttons              (EditColorRow      *self);

static void
color_button_clicked (GtkWidget    *button,
                        EditColorRow *self)
{
  GdkRGBA *color;

  color = g_object_get_data (G_OBJECT (button), "color");
  gtk_popover_popdown (GTK_POPOVER (self->color_popover));

  g_signal_emit (self, obj_signals[CHANGED], 0, color);
}

static GtkButton *
set_up_color_button (EditColorRow    *self,
                     GtkWidget       *box,
                     guint            color_index)
{
  GtkWidget *swatch;
  const GdkRGBA *rgba;

  rgba = &style_color_palette[color_index];
  swatch = g_object_new (EDIT_TYPE_COLOR_SWATCH,
                        "color", rgba,
                         NULL);
  self->swatches [color_index] = swatch;

  g_object_set_data (G_OBJECT (swatch), "color", (gpointer) rgba);

  gtk_box_append (GTK_BOX (box), swatch);

  g_signal_connect (swatch, "clicked", G_CALLBACK (color_button_clicked), self);

  return GTK_BUTTON (swatch);
}

/* We call this when the dark mode changes to recaluclate the CSS
 * style of the swatches. This lets us batch that call instead of
 * adding 100s of signal handlers for each widget.
 *
 * The reason we recalculate is the contrast-color of the swatch is
 * pulled from the theme and switches from light-to-dark when the the
 * mode changes, but we need a fixed light/dark setting based on the
 * luminocity of the swatch.
 */
static void
recalculate_dark_mode (EditColorRow *self)
{
  for (guint i = 0; i < G_N_ELEMENTS (style_color_palette); i ++)
    edit_color_swatch_recalculate_style (EDIT_COLOR_SWATCH (self->swatches[i]));
}

static void
edit_color_row_init (EditColorRow *self)
{
  GtkWidget *box = NULL;

  gtk_widget_init_template (GTK_WIDGET (self));

  for (guint i = 0; i < G_N_ELEMENTS (style_color_palette); i ++)
    {
      if (i % STYLE_COLOR_PALETTE_STRIDE == 0)
        {
          box = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
          gtk_box_append (GTK_BOX (self->palette_box), box);
          gtk_widget_add_css_class (box, "linked");
        }
      set_up_color_button (self, box, i);
    }
  self->color_set = FALSE;

  self->style_recalculate_handler =
    g_signal_connect_swapped (adw_style_manager_get_default (),
                              "notify::dark",
                              G_CALLBACK (recalculate_dark_mode),
                              self);
}

static void
edit_color_row_class_init (EditColorRowClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
  AdwActionRowClass *action_row_class = ADW_ACTION_ROW_CLASS (klass);

  object_class->set_property = edit_color_row_set_property;
  object_class->get_property = edit_color_row_get_property;
  object_class->finalize = edit_color_row_finalize;

  action_row_class->activate = edit_color_row_activate;

  gtk_widget_class_set_template_from_resource (widget_class, "/org/gnome/Crosswords/edit-color-row.ui");

  gtk_widget_class_bind_template_child (widget_class, EditColorRow, color_popover);
  gtk_widget_class_bind_template_child (widget_class, EditColorRow, palette_box);
  gtk_widget_class_bind_template_child (widget_class, EditColorRow, clear_button);
  gtk_widget_class_bind_template_callback (widget_class, notify_popover_visible_cb);
  gtk_widget_class_bind_template_callback (widget_class, clear_button_cb);

  gtk_widget_class_set_css_name (widget_class, "row");

  obj_signals[CHANGED] =
    g_signal_new ("changed",
                  EDIT_TYPE_COLOR_ROW,
                  G_SIGNAL_RUN_FIRST,
                  0,
                  NULL, NULL,
                  NULL,
                  G_TYPE_NONE, 1,
                  GDK_TYPE_RGBA);
  obj_signals[CLEARED] =
    g_signal_new ("cleared",
                  EDIT_TYPE_COLOR_ROW,
                  G_SIGNAL_RUN_FIRST,
                  0,
                  NULL, NULL,
                  NULL,
                  G_TYPE_NONE, 0);

  obj_props [PROP_CURRENT_COLOR] = g_param_spec_boxed ("current-color",
                                                       NULL, NULL,
                                                       GDK_TYPE_RGBA,
                                                       G_PARAM_READWRITE);
  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
edit_color_row_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *psec)
{
  EditColorRow *self;

  g_return_if_fail (EDIT_IS_COLOR_ROW (object));

  self = EDIT_COLOR_ROW (object);

  switch (prop_id)
    {
      case PROP_CURRENT_COLOR:
        if (g_value_get_boxed (value))
          {
            self->current_color = *(GdkRGBA *) g_value_get_boxed (value);
            self->color_set = TRUE;
          }
        else
          {
            self->color_set = FALSE;
          }
        update_buttons (self);
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, psec);
        break;
    }
}

static void
edit_color_row_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *psec)
{
  EditColorRow *self;

  g_return_if_fail (EDIT_IS_COLOR_ROW (object));

  self = EDIT_COLOR_ROW (object);

  switch (prop_id)
    {
      case PROP_CURRENT_COLOR:
        g_value_set_boxed (value, &self->current_color);
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, psec);
        break;
    }
}

static void
edit_color_row_finalize (GObject *object)
{
  EditColorRow *self;

  self = EDIT_COLOR_ROW (object);

  g_clear_signal_handler (&self->style_recalculate_handler, adw_style_manager_get_default());
}

static void
edit_color_row_activate (AdwActionRow *action_row)
{
  EditColorRow *self;

  self = EDIT_COLOR_ROW (action_row);

  gtk_popover_popup (GTK_POPOVER (self->color_popover));
}

static void
notify_popover_visible_cb (EditColorRow *self)
{
  /* this prevents clicks on the popover widget also appearing visually on the row */
  if (gtk_widget_get_visible (GTK_WIDGET (self->color_popover)))
    gtk_widget_add_css_class (GTK_WIDGET (self), "has-open-popup");
  else
    gtk_widget_remove_css_class (GTK_WIDGET (self), "has-open-popup");
}

static void
clear_button_cb (EditColorRow *self)
{
  gtk_popover_popdown (GTK_POPOVER (self->color_popover));

  g_signal_emit (self, obj_signals[CLEARED], 0);
}


static void
update_buttons (EditColorRow *self)
{
  const GdkRGBA *rgba;
  GtkWidget *swatch;

  for (guint i = 0; i < G_N_ELEMENTS (style_color_palette); i++)
    {
      rgba = &style_color_palette[i];
      swatch = self->swatches[i];

      g_signal_handlers_block_by_func (swatch, color_button_clicked, self);

      if (self->color_set && gdk_rgba_equal (rgba, &self->current_color))
        edit_color_swatch_set_selected (EDIT_COLOR_SWATCH (swatch), TRUE);
      else
        edit_color_swatch_set_selected (EDIT_COLOR_SWATCH (swatch), FALSE);

      g_signal_handlers_unblock_by_func (swatch, color_button_clicked, self);
    }
}
