/* edit-color-swatch.c
 *
 * Copyright 2024 Jonathan Blandford
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#include "crosswords-config.h"
#include <glib/gi18n-lib.h>
#include "color-palette.h"
#include "crosswords-misc.h"
#include "edit-color-swatch.h"


enum
{
  PROP_0,
  PROP_COLOR,
  PROP_SELECTED,
  N_PROPS
};

static GParamSpec *obj_props[N_PROPS] = {0, };


struct _EditColorSwatch
{
  GtkButton parent_instance;

  GtkWidget *icon_stack;

  GdkRGBA color;
  gboolean selected;
};


G_DEFINE_FINAL_TYPE (EditColorSwatch, edit_color_swatch, GTK_TYPE_BUTTON);


static void edit_color_swatch_init         (EditColorSwatch      *self);
static void edit_color_swatch_class_init   (EditColorSwatchClass *klass);
static void edit_color_swatch_set_property (GObject              *object,
                                            guint                 prop_id,
                                            const GValue         *value,
                                            GParamSpec           *psec);
static void edit_color_swatch_get_property (GObject              *object,
                                            guint                 prop_id,
                                            GValue               *value,
                                            GParamSpec           *psec);
static void edit_color_swatch_snapshot     (GtkWidget            *widget,
                                            GtkSnapshot          *snapshot);


static void
edit_color_swatch_init (EditColorSwatch *self)
{
  gtk_widget_init_template (GTK_WIDGET (self));
}

static void
edit_color_swatch_class_init (EditColorSwatchClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);

  object_class->set_property = edit_color_swatch_set_property;
  object_class->get_property = edit_color_swatch_get_property;
  widget_class->snapshot = edit_color_swatch_snapshot;

  gtk_widget_class_set_template_from_resource (widget_class,
                                               "/org/gnome/Crosswords/edit-color-swatch.ui");

  gtk_widget_class_bind_template_child (widget_class, EditColorSwatch, icon_stack);

  obj_props [PROP_COLOR] = g_param_spec_boxed ("color",
                                               NULL, NULL,
                                               GDK_TYPE_RGBA,
                                               G_PARAM_READWRITE);

  obj_props [PROP_SELECTED] = g_param_spec_boolean ("selected",
                                                    NULL, NULL,
                                                    FALSE,
                                                    G_PARAM_READWRITE);

  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
edit_color_swatch_set_property (GObject              *object,
                                guint                 prop_id,
                                const GValue         *value,
                                GParamSpec           *psec)
{
  EditColorSwatch *self;

  g_return_if_fail (EDIT_IS_COLOR_SWATCH (object));

  self = EDIT_COLOR_SWATCH (object);

  switch (prop_id)
    {
    case PROP_COLOR:
      edit_color_swatch_set_color (self, (GdkRGBA *)g_value_get_boxed (value));
      break;
    case PROP_SELECTED:
      edit_color_swatch_set_selected (self, g_value_get_boolean (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, psec);
      break;
    }
}

static void
edit_color_swatch_get_property (GObject              *object,
                                guint                 prop_id,
                                GValue               *value,
                                GParamSpec           *psec)
{
  EditColorSwatch *self;

  g_return_if_fail (EDIT_IS_COLOR_SWATCH (object));

  self = EDIT_COLOR_SWATCH (object);

  switch (prop_id)
    {
    case PROP_COLOR:
      g_value_set_boxed (value, &self->color);
      break;
    case PROP_SELECTED:
      g_value_set_boolean (value, self->selected);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, psec);
      break;
    }
}

static void
edit_color_swatch_snapshot (GtkWidget   *widget,
                            GtkSnapshot *snapshot)
{
  EditColorSwatch *self;
  gfloat width, height;

  self = EDIT_COLOR_SWATCH (widget);

  width = (gfloat) gtk_widget_get_width (widget);
  height = (gfloat) gtk_widget_get_height (widget);

  gtk_snapshot_append_color (snapshot, &self->color,
                             &GRAPHENE_RECT_INIT (0.0, 0.0, width, height));

  GTK_WIDGET_CLASS (edit_color_swatch_parent_class)->snapshot (widget, snapshot);
}

void
edit_color_swatch_recalculate_style (EditColorSwatch *self)
{
  gboolean light;

  g_return_if_fail (EDIT_IS_COLOR_SWATCH (self));

  light = (PALETTE_COLOR_LUMINANCE (&self->color) > 1.6);

  /* If we're in dark mode, we need to invert this, as the meaning of
   * 'light' is inverted */
  if (render_color_scheme_from_adwaita () == RENDER_COLOR_SCHEME_DARK)
    light = !light;

  if (light)
    gtk_widget_add_css_class (GTK_WIDGET (self), "light");
  else
    gtk_widget_remove_css_class (GTK_WIDGET (self), "light");

  gtk_widget_queue_draw (GTK_WIDGET (self));
}

void
edit_color_swatch_set_color (EditColorSwatch *self,
                             GdkRGBA         *color)
{
  g_return_if_fail (EDIT_IS_COLOR_SWATCH (self));
  g_return_if_fail (color != NULL);

  self->color = *color;

  edit_color_swatch_recalculate_style (self);
}

void
edit_color_swatch_set_selected (EditColorSwatch *self,
                                gboolean         selected)
{
  g_return_if_fail (EDIT_IS_COLOR_SWATCH (self));

  selected = !!selected;

  if (selected != self->selected)
    {
      self->selected = selected;

      if (self->selected)
        gtk_stack_set_visible_child_name (GTK_STACK (self->icon_stack),
                                          "selected");
      else
        gtk_stack_set_visible_child_name (GTK_STACK (self->icon_stack),
                                          "empty");

      g_object_notify_by_pspec (G_OBJECT (self), obj_props [PROP_SELECTED]);
    }
}

gboolean
edit_color_swatch_get_selected (EditColorSwatch *self,
                                gboolean         selected)
{
  g_return_val_if_fail (EDIT_IS_COLOR_SWATCH (self), FALSE);

  return self->selected;
}
