/* crosswords-thumbnailer.c
 *
 * Copyright 2023 Tanmay Patil <tanmaynpatil105@gmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#include "crosswords-config.h"
#include "puzzle-render.h"

#include <locale.h>

#include <libipuz/libipuz.h>
#include <cairo-svg.h>
#include <glib/gi18n-lib.h>

#define THUMBNAIL_SIZE 128

static gint size = THUMBNAIL_SIZE;
static gboolean export_svg = FALSE;
static gboolean dark_scheme = FALSE;
static const gchar **file_arguments;

static const GOptionEntry goption_options[] = {
  { "size", 's', 0, G_OPTION_ARG_INT, &size, NULL, "SIZE" },
  { "svg",  'v', 0, G_OPTION_ARG_NONE, &export_svg, NULL, NULL},
  { "dark", 'd', 0, G_OPTION_ARG_NONE, &dark_scheme, NULL, NULL},
  { G_OPTION_REMAINING, 0, 0, G_OPTION_ARG_FILENAME_ARRAY, &file_arguments, NULL, "<input> <ouput>" },
  { NULL, 0, 0, 0, NULL, NULL, NULL }
};

static void
print_usage (GOptionContext *context)
{
  g_autofree gchar* help = NULL;

  help = g_option_context_get_help (context, TRUE, NULL);
  g_print ("%s", help);
}

static cairo_status_t
crosswords_thumbnail_get (IpuzPuzzle  *puzzle,
                          const gchar *output)
{
  cairo_surface_t *surf;
  cairo_t *cr;
  RenderViewport viewport = { 0.0, 0.0, size, size};
  LayoutConfig config = {
    .border_size = 1,
    .base_size = 3,
  };
  cairo_status_t status;

  if (export_svg)
    surf = cairo_svg_surface_create (output, size, size);
  else
    surf = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, size, size);

  cr = cairo_create (surf);
  render_puzzle_with_config (cr,
                             &viewport,
                             puzzle,
                             dark_scheme ? RENDER_COLOR_SCHEME_DARK : RENDER_COLOR_SCHEME_LIGHT,
                             config);
  status = cairo_status (cr);

  if (status == CAIRO_STATUS_SUCCESS)
    {
      if (export_svg)
        cairo_surface_flush(surf);
      else
        status = cairo_surface_write_to_png (surf, output);
    }

  cairo_destroy (cr);
  cairo_surface_destroy (surf);

  return status;
}

int
main (int   argc,
      char *argv[])
{
  g_autoptr (IpuzPuzzle) puzzle = NULL;
  g_autoptr (GError)  error = NULL;
  g_autoptr (GOptionContext) context = NULL;
  const gchar *input;
  const gchar *output;
  cairo_status_t status;

  setlocale (LC_ALL, "");

  context = g_option_context_new ("- GNOME Crosswords Thumbnailer");
  g_option_context_add_main_entries (context, goption_options, NULL);

  if (!g_option_context_parse (context, &argc, &argv, &error))
    {
      g_printerr ("%s\n", error->message);
      print_usage (context);
      return -1;
    }

  input = file_arguments ? file_arguments[0] : NULL;
  output = input ? file_arguments[1] : NULL;

  if (!input || !output)
    {
      print_usage (context);
      return -1;
    }

  puzzle = ipuz_puzzle_new_from_file (input, &error);
  if (error != NULL)
    {
      g_printerr ("Error loading puzzle: %s\n", error->message);
      return -1;
    }

  if ((status = crosswords_thumbnail_get (puzzle, output)) != CAIRO_STATUS_SUCCESS)
    {
      g_printerr ("Error creating thumbnail: %s\n", cairo_status_to_string (status));
      return -1;
    }

  return 0;
}
